"""Density compensation functions.

"""
import sigpy as sp
from tqdm.auto import tqdm


__all__ = ['pipe_menon_dcf']


def pipe_menon_dcf(coord, img_shape=None, device=sp.cpu_device, max_iter=30,
                   n=128, beta=8, width=4, show_pbar=True):
    r"""Compute Pipe Menon density compensation factor.

    Perform the following iteration:

    .. math::

        w = \frac{w}{|G^H G w|}

    with :math:`G` as the gridding operator.

    Args:
        coord (array): k-space coordinates.
        img_shape (None or list): Image shape.
        device (Device): computing device.
        max_iter (int): number of iterations.
        n (int): Kaiser-Bessel sampling numbers for gridding operator.
        beta (float): Kaiser-Bessel kernel parameter.
        width (float): Kaiser-Bessel kernel width.
        show_pbar (bool): show progress bar.

    Returns:
        array: density compensation factor.

    References:
        Pipe, James G., and Padmanabhan Menon.
        Sampling Density Compensation in MRI:
        Rationale and an Iterative Numerical Solution.
        Magnetic Resonance in Medicine 41, no. 1 (1999): 179–86.


    """
    device = sp.Device(device)
    xp = device.xp

    with device:
        w = xp.ones(coord.shape[:-1], dtype=coord.dtype)
        if img_shape is None:
            img_shape = sp.estimate_shape(coord)

        G = sp.linop.Gridding(img_shape, coord, param=beta,
                              width=width, kernel='kaiser_bessel')
        with tqdm(total=max_iter, desc="PipeMenonDCF",
                  disable=not show_pbar) as pbar:
            for it in range(max_iter):
                GHGw = G.H * G * w
                w /= xp.abs(GHGw)
                resid = xp.abs(GHGw - 1).max().item()

                pbar.set_postfix(resid='{0:.2E}'.format(resid))
                pbar.update()

    return w



"""The module contains functions and classes for MRI reconstruction.

It provides convenient simulation and sampling functions,
such as the poisson-disc sampling function. It also
provides functions to compute preconditioners,
and density compensation factors.

"""
from sigpy.mri import app, linop

from sigpy.mri import dcf, precond, samp, sim, util
from sigpy.mri.dcf import *  # noqa
from sigpy.mri.precond import *  # noqa
from sigpy.mri.samp import *  # noqa
from sigpy.mri.sim import *  # noqa
from sigpy.mri.util import *  # noqa

__all__ = ['app', 'linop']
__all__.extend(dcf.__all__)
__all__.extend(precond.__all__)
__all__.extend(samp.__all__)
__all__.extend(sim.__all__)
__all__.extend(util.__all__)



# -*- coding: utf-8 -*-
"""Convolution functions with  multi-dimension, and multi-channel support.

"""
import numpy as np
import scipy.signal as signal
from sigpy import backend, util, config


__all__ = ['convolve', 'convolve_data_adjoint', 'convolve_filter_adjoint']


def convolve(data, filt, mode='full', strides=None, multi_channel=False):
    r"""Convolution that supports multi-dimensional and multi-channel inputs.

    This function follows the signal processing definition of convolution.

    Args:
        data (array): data array of shape:
            :math:`[..., m_1, ..., m_D]` if multi_channel is False,
            :math:`[..., c_i, m_1, ..., m_D]` otherwise.
        filt (array): filter array of shape:
            :math:`[n_1, ..., n_D]` if multi_channel is False
            :math:`[c_o, c_i, n_1, ..., n_D]` otherwise.
        mode (str): {'full', 'valid'}.
        strides (None or tuple of ints): convolution strides of length D.
        multi_channel (bool): specify if input/output has multiple channels.

    Returns:
        array: output array of shape:
            :math:`[..., p_1, ..., p_D]` if multi_channel is False,
            :math:`[..., c_o, p_1, ..., p_D]` otherwise.

    """
    xp = backend.get_array_module(data)
    if xp == np:
        output = _convolve(data, filt, mode=mode, strides=strides,
                           multi_channel=multi_channel)
    else:  # pragma: no cover
        if config.cudnn_enabled:
            if np.issubdtype(data.dtype, np.floating):
                output = _convolve_cuda(data, filt,
                                        mode=mode, strides=strides,
                                        multi_channel=multi_channel)
            else:
                output = _complex(_convolve_cuda, data, filt,
                                  mode=mode, strides=strides,
                                  multi_channel=multi_channel)
        else:
            raise RuntimeError(
                'cudnn must be installed to perform convolution on GPU.')

    return output


def convolve_data_adjoint(output, filt, data_shape,
                          mode='full', strides=None, multi_channel=False):
    """Adjoint convolution operation with respect to data.

    Args:
        output (array): output array of shape
            :math:`[..., p_1, ..., p_D]` if multi_channel is False,
            :math:`[..., c_o, p_1, ..., p_D]` otherwise.
        filt (array): filter array of shape
            :math:`[n_1, ..., n_D]` if multi_channel is False
            :math:`[c_o, c_i, n_1, ..., n_D]` otherwise.
        mode (str): {'full', 'valid'}.
        strides (None or tuple of ints): convolution strides of length D.
        multi_channel (bool): specify if input/output has multiple channels.
        multi_channel (bool): specify if data/output has multiple channels.
        mode (str): {'full', 'valid'}.

    Returns:
        array: data array of shape
            :math:`[..., m_1, ..., m_D]` if multi_channel is False,
            :math:`[..., c_i, m_1, ..., m_D]` otherwise.

    """
    data_shape = tuple(data_shape)

    xp = backend.get_array_module(output)
    if xp == np:
        data = _convolve_data_adjoint(output, filt, data_shape,
                                      mode=mode, strides=strides,
                                      multi_channel=multi_channel)
    else:  # pragma: no cover
        if config.cudnn_enabled:
            if np.issubdtype(output.dtype, np.floating):
                data = _convolve_data_adjoint_cuda(
                    output, filt, data_shape,
                    mode=mode, strides=strides,
                    multi_channel=multi_channel)
            else:
                data = _complex(_convolve_data_adjoint_cuda,
                                output, filt.conj(), data_shape,
                                mode=mode, strides=strides,
                                multi_channel=multi_channel)
        else:
            raise RuntimeError(
                'cudnn must be installed to perform convolution on GPU.')

    return data


def convolve_filter_adjoint(output, data, filt_shape,
                            mode='full', strides=None, multi_channel=False):
    """Adjoint convolution operation with respect to filter.

    Args:
        output (array): output array of shape:
            :math:`[..., p_1, ..., p_D]` if multi_channel is False,
            :math:`[..., c_o, p_1, ..., p_D]` otherwise.
        data (array): data array of shape:
            :math:`[..., m_1, ..., m_D]` if multi_channel is False,
            :math:`[..., c_i, m_1, ..., m_D]` otherwise.
        mode (str): {'full', 'valid'}.
        strides (None or tuple of ints): convolution strides of length D.
        multi_channel (bool): specify if input/output has multiple channels.

    Returns:
        array: filter array of shape:
            :math:`[n_1, ..., n_D]` if multi_channel is False
            :math:`[c_o, c_i, n_1, ..., n_D]` otherwise.

    """
    filt_shape = tuple(filt_shape)
    xp = backend.get_array_module(data)
    if xp == np:
        filt = _convolve_filter_adjoint(output, data, filt_shape,
                                        mode=mode, strides=strides,
                                        multi_channel=multi_channel)
    else:  # pragma: no cover
        if config.cudnn_enabled:
            if np.issubdtype(output.dtype, np.floating):
                filt = _convolve_filter_adjoint_cuda(
                    output, data, filt_shape,
                    mode=mode, strides=strides,
                    multi_channel=multi_channel)
            else:
                filt = _complex(_convolve_filter_adjoint_cuda,
                                output, data.conj(), filt_shape,
                                mode=mode, strides=strides,
                                multi_channel=multi_channel)
        else:
            raise RuntimeError(
                'cudnn must be installed to perform convolution on GPU.')

    return filt


def _get_convolve_params(data_shape, filt_shape,
                         mode, strides, multi_channel):
    D = len(filt_shape) - 2 * multi_channel
    m = tuple(data_shape[-D:])
    n = tuple(filt_shape[-D:])
    b = tuple(data_shape[:-D - multi_channel])
    B = util.prod(b)

    if multi_channel:
        if filt_shape[-D - 1] != data_shape[-D - 1]:
            raise ValueError('Data channel mismatch, '
                             'got {} from data and {} from filt.'.format(
                                 data_shape[-D - 1], filt_shape[-D - 1]))

        c_i = filt_shape[-D - 1]
        c_o = filt_shape[-D - 2]
    else:
        c_i = 1
        c_o = 1

    if strides is None:
        s = (1, ) * D
    else:
        if len(strides) != D:
            raise ValueError('Strides must have length {}.'.format(D))

        s = tuple(strides)

    if mode == 'full':
        p = tuple((m_d + n_d - 1 + s_d - 1) // s_d
                  for m_d, n_d, s_d in zip(m, n, s))
    elif mode == 'valid':
        if (any(m_d >= n_d for m_d, n_d in zip(m, n)) and
            any(m_d < n_d for m_d, n_d in zip(m, n))):
            raise ValueError('In valid mode, either data or filter must be '
                             'at least as large as the other in every axis.')

        p = tuple((m_d - n_d + 1 + s_d - 1) // s_d
                  for m_d, n_d, s_d in zip(m, n, s))
    else:
        raise ValueError('Invalid mode, got {}'.format(mode))

    return D, b, B, m, n, s, c_i, c_o, p


def _convolve(data, filt,
              mode='full', strides=None,
              multi_channel=False):
    D, b, B, m, n, s, c_i, c_o, p = _get_convolve_params(
        data.shape, filt.shape,
        mode, strides, multi_channel)

    # Normalize shapes.
    data = data.reshape((B, c_i) + m)
    filt = filt.reshape((c_o, c_i) + n)
    output = np.zeros((B, c_o) + p, dtype=data.dtype)
    slc = tuple(slice(None, None, s_d) for s_d in s)

    for k in range(B):
        for j in range(c_o):
            for i in range(c_i):
                output[k, j] += signal.convolve(
                    data[k, i], filt[j, i], mode=mode)[slc]

    # Reshape.
    if multi_channel:
        output = output.reshape(b + (c_o, ) + p)
    else:
        output = output.reshape(b + p)

    return output


def _convolve_data_adjoint(output, filt, data_shape,
                           mode='full', strides=None,
                           multi_channel=False):
    D, b, B, m, n, s, c_i, c_o, p = _get_convolve_params(
        data_shape, filt.shape,
        mode, strides, multi_channel)

    # Normalize shapes.
    output = output.reshape((B, c_o) + p)
    filt = filt.reshape((c_o, c_i) + n)
    data = np.zeros((B, c_i) + m, dtype=output.dtype)
    slc = tuple(slice(None, None, s_d) for s_d in s)
    if mode == 'full':
        output_kj = np.zeros([m_d + n_d - 1 for m_d, n_d in zip(m, n)],
                             dtype=output.dtype)
        adjoint_mode = 'valid'
    elif mode == 'valid':
        output_kj = np.zeros([max(m_d, n_d) - min(m_d, n_d) + 1
                              for m_d, n_d in zip(m, n)],
                             dtype=output.dtype)
        if all(m_d >= n_d for m_d, n_d in zip(m, n)):
            adjoint_mode = 'full'
        else:
            adjoint_mode = 'valid'

    for k in range(B):
        for j in range(c_o):
            for i in range(c_i):
                output_kj[slc] = output[k, j]
                data[k, i] += signal.correlate(
                    output_kj, filt[j, i],
                    mode=adjoint_mode)

    # Reshape.
    data = data.reshape(data_shape)
    return data


def _convolve_filter_adjoint(output, data, filt_shape,
                             mode='full', strides=None,
                             multi_channel=False):
    D, b, B, m, n, s, c_i, c_o, p = _get_convolve_params(
        data.shape, filt_shape,
        mode, strides, multi_channel)

    # Normalize shapes.
    data = data.reshape((B, c_i) + m)
    output = output.reshape((B, c_o) + p)
    slc = tuple(slice(None, None, s_d) for s_d in s)
    if mode == 'full':
        output_kj = np.zeros([m_d + n_d - 1 for m_d, n_d in zip(m, n)],
                             dtype=output.dtype)
        adjoint_mode = 'valid'
    elif mode == 'valid':
        output_kj = np.zeros([max(m_d, n_d) - min(m_d, n_d) + 1
                              for m_d, n_d in zip(m, n)],
                             dtype=output.dtype)
        if all(m_d >= n_d for m_d, n_d in zip(m, n)):
            adjoint_mode = 'valid'
        else:
            adjoint_mode = 'full'

    filt = np.zeros((c_o, c_i) + n, dtype=output.dtype)
    for k in range(B):
        for j in range(c_o):
            for i in range(c_i):
                output_kj[slc] = output[k, j]
                filt[j, i] += signal.correlate(
                    output_kj, data[k, i], mode=adjoint_mode)

    # Reshape.
    filt = filt.reshape(filt_shape)
    return filt


if config.cudnn_enabled:  # pragma: no cover
    from cupy import cudnn

    def _complex(func, data1, data2, *kargs, **kwargs):
        """Helper function to convert func to support complex floats.
        """
        xp = backend.get_array_module(data1)
        data1r = xp.real(data1)
        data1i = xp.imag(data1)
        data2r = xp.real(data2)
        data2i = xp.imag(data2)

        outputr = func(data1r, data2r, *kargs, **kwargs)
        outputr -= func(data1i, data2i, *kargs, **kwargs)
        outputi = func(data1i, data2r, *kargs, **kwargs)
        outputi += func(data1r, data2i, *kargs, **kwargs)

        output = outputr + 1j * outputi
        output = output.astype(data1.dtype, copy=False)
        return output

    def _convolve_cuda(data, filt,
                       mode='full', strides=None,
                       multi_channel=False):
        xp = backend.get_array_module(data)

        D, b, B, m, n, s, c_i, c_o, p = _get_convolve_params(
            data.shape, filt.shape,
            mode, strides, multi_channel)
        dilations = (1, ) * D
        groups = 1
        auto_tune = True
        tensor_core = 'auto'
        if mode == 'full':
            pads = tuple(n_d - 1 for n_d in n)
        elif mode == 'valid':
            pads = (0, ) * D

        data = data.reshape((B, c_i) + m)
        filt = filt.reshape((c_o, c_i) + n)
        output = xp.empty((B, c_o) + p, dtype=data.dtype)
        filt = util.flip(filt, axes=range(-D, 0))
        cudnn.convolution_forward(data, filt, None, output,
                                  pads, s, dilations, groups,
                                  auto_tune=auto_tune,
                                  tensor_core=tensor_core)

        # Reshape.
        if multi_channel:
            output = output.reshape(b + (c_o, ) + p)
        else:
            output = output.reshape(b + p)

        return output

    def _convolve_data_adjoint_cuda(output, filt, data_shape,
                                    mode='full', strides=None,
                                    multi_channel=False):
        xp = backend.get_array_module(output)

        D, b, B, m, n, s, c_i, c_o, p = _get_convolve_params(
            data_shape, filt.shape,
            mode, strides, multi_channel)
        dilations = (1, ) * D
        groups = 1
        auto_tune = True
        tensor_core = 'auto'
        deterministic = False
        if mode == 'full':
            pads = tuple(n_d - 1 for n_d in n)
        elif mode == 'valid':
            pads = (0, ) * D

        output = output.reshape((B, c_o) + p)
        filt = filt.reshape((c_o, c_i) + n)
        data = xp.empty((B, c_i) + m, dtype=output.dtype)
        filt = util.flip(filt, axes=range(-D, 0))
        cudnn.convolution_backward_data(filt, output, None, data,
                                        pads, s, dilations, groups,
                                        deterministic=deterministic,
                                        auto_tune=auto_tune,
                                        tensor_core=tensor_core)

        # Reshape.
        data = data.reshape(data_shape)

        return data

    def _convolve_filter_adjoint_cuda(output, data, filt_shape,
                                      mode='full', strides=None,
                                      multi_channel=False):
        xp = backend.get_array_module(data)

        D, b, B, m, n, s, c_i, c_o, p = _get_convolve_params(
            data.shape, filt_shape,
            mode, strides, multi_channel)
        dilations = (1, ) * D
        groups = 1
        auto_tune = True
        tensor_core = 'auto'
        deterministic = False
        if mode == 'full':
            pads = tuple(n_d - 1 for n_d in n)
        elif mode == 'valid':
            pads = (0, ) * D

        data = data.reshape((B, c_i) + m)
        output = output.reshape((B, c_o) + p)
        filt = xp.empty((c_o, c_i) + n, dtype=output.dtype)
        cudnn.convolution_backward_filter(data, output, filt,
                                          pads, s, dilations, groups,
                                          deterministic=deterministic,
                                          auto_tune=auto_tune,
                                          tensor_core=tensor_core)
        filt = util.flip(filt, axes=range(-D, 0))
        filt = filt.reshape(filt_shape)

        return filt



# -*- coding: utf-8 -*-
"""Configuration.

This module contains flags to turn on and off optional modules.

"""
import warnings
from importlib import util

cupy_enabled = util.find_spec("cupy") is not None
if cupy_enabled:
    try:
        import cupy  # noqa
    except ImportError as e:
        warnings.warn(
            f"Importing cupy failed. "
            f"For more details, see the error stack below:\n{e}"
        )
        cupy_enabled = False

if cupy_enabled:  # pragma: no cover
    cudnn_enabled = util.find_spec("cupy.cuda.cudnn") is not None
    nccl_enabled = util.find_spec("cupy.cuda.nccl") is not None
else:
    cudnn_enabled = False
    nccl_enabled = False

mpi4py_enabled = util.find_spec("mpi4py") is not None

# This is to catch an import error when the cudnn in cupy (system) and pytorch
# (built in) are in conflict.
if util.find_spec("torch") is not None:
    try:
        import torch  # noqa
        pytorch_enabled = True
    except ImportError:
        print('Warning : Pytorch installed but can import')
        pytorch_enabled = False
else:
    pytorch_enabled = False



# -*- coding: utf-8 -*-
"""Multiband RF Pulse Design functions.
"""
import numpy as np

__all__ = ['mb_phs_tab', 'mb_rf', 'dz_pins']

from sigpy.mri.rf.trajgrad import trap_grad
from sigpy.mri.rf import slr as slr


def mb_rf(pulse_in, n_bands=3, band_sep=5, phs_0_pt='None'):
    r"""Multiband an input RF pulse.

     Args:
         pulse_in (array): samples of single-band RF pulse.
         n_bands (int): number of bands.
         band_sep (float): normalized slice separation.
         phs_0_pt (string): set of phases to use. Can be 'phs_mod' (Wong),
            'amp_mod' (Malik), 'quad_mod' (Grissom), or 'None'

     band_sep = slice_sep/slice_thick*tb, where tb is time-bandwidth product
     of the single-band pulse

     Returns:
         array: multibanded pulse out

     References:
         Wong, E. (2012). 'Optimized Phase Schedules for Minimizing Peak RF
         Power in Simultaneous Multi-Slice RF Excitation Pulses'. Proc. Intl.
         Soc. Mag. Reson. Med., 20 p. 2209.
         Malik, S. J., Price, A. N., and Hajnal, J. V. (2015). 'Optimized
         Amplitude Modulated Multi-Band RF pulses'. Proc. Intl. Soc. Mag.
         Reson. Med., 23 p. 2398.
     """

    if phs_0_pt != 'None':
        phs = mb_phs_tab(n_bands, phs_0_pt)
    else:
        phs = np.zeros(n_bands)

    # build multiband modulation function
    n = np.size(pulse_in)
    b = np.zeros(n, dtype='complex')
    for ii in range(0, n_bands):
        b += np.exp(1j * 2 * np.pi / n * band_sep * np.arange(-n / 2, n / 2, 1)
                    * (ii - (n_bands - 1) / 2)) * np.exp(1j * phs[ii])

    pulse_out = b * pulse_in

    return pulse_out


def mb_phs_tab(n_bands, phs_type='phs_mod'):
    # Return phases to minimize peak b1 amplitude of an MB pulse

    if phs_type == 'phs_mod':

        if n_bands < 3 or n_bands > 16:
            raise Exception('Wongs phases valid for 2 < nBands < 17.')

        # Eric Wong's phases: From E C Wong, ISMRM 2012, p. 2209
        p = np.zeros((14, 16))
        p[0, 1:3] = np.array([0.73, 4.602])
        p[1, 1:4] = np.array([3.875, 5.94, 6.197])
        p[2, 1:5] = np.array([3.778, 5.335, 0.872, 0.471])
        p[3, 1:6] = np.array([2.005, 1.674, 5.012, 5.736, 4.123])
        p[4, 1:7] = np.array([3.002, 5.998, 5.909, 2.624, 2.528, 2.440])
        p[5, 1:8] = np.array([1.036, 3.414, 3.778, 3.215, 1.756, 4.555, 2.467])
        p[6, 1:9] = np.array([1.250, 1.783, 3.558, 0.739, 3.319, 1.296,
                              0.521, 5.332])
        p[7, 1:10] = np.array([4.418, 2.360, 0.677, 2.253, 3.472, 3.040,
                               3.974, 1.192, 2.510])
        p[8, 1:11] = np.array([5.041, 4.285, 3.001, 5.765, 4.295, 0.056,
                               4.213, 6.040, 1.078, 2.759])
        p[9, 1:12] = np.array([2.755, 5.491, 4.447, 0.231, 2.499, 3.539,
                               2.931, 2.759, 5.376, 4.554, 3.479])
        p[10, 1:13] = np.array([0.603, 0.009, 4.179, 4.361, 4.837, 0.816,
                                5.995, 4.150, 0.417, 1.520, 4.517, 1.729])
        p[11, 1:14] = np.array([3.997, 0.830, 5.712, 3.838, 0.084, 1.685,
                                5.328, 0.237, 0.506, 1.356, 4.025, 4.483,
                                4.084])
        p[12, 1:15] = np.array([4.126, 2.266, 0.957, 4.603, 0.815, 3.475,
                                0.977, 1.449, 1.192, 0.148, 0.939, 2.531,
                                3.612, 4.801])
        p[13, 1:16] = np.array([4.359, 3.510, 4.410, 1.750, 3.357, 2.061,
                                5.948, 3.000, 2.822, 0.627, 2.768, 3.875,
                                4.173, 4.224, 5.941])

        out = p[n_bands - 3, 0:n_bands]

    elif phs_type == 'amp_mod':

        # Malik's Hermitian phases: From S J Malik, ISMRM 2015, p. 2398
        if n_bands < 4 or n_bands > 12:
            raise Exception('Maliks phases valid for 3 < nBands < 13.')

        p = np.zeros((9, 12))
        p[0, 0:4] = np.array([0, np.pi, np.pi, 0])
        p[1, 0:5] = np.array([0, 0, np.pi, 0, 0])
        p[2, 0:6] = np.array([1.691, 2.812, 1.157, -1.157, -2.812, -1.691])
        p[3, 0:7] = np.array([2.582, -0.562, 0.102, 0, -0.102, 0.562, -2.582])
        p[4, 0:8] = np.array([2.112, 0.220, 1.464, 1.992, -1.992, -1.464,
                              -0.220, -2.112])
        p[5, 0:9] = np.array([0.479, -2.667, -0.646, -0.419, 0, 0.419, 0.646,
                              2.667, -0.479])
        p[6, 0:10] = np.array([1.683, -2.395, 2.913, 0.304, 0.737, -0.737,
                               -0.304, -2.913, 2.395, -1.683])
        p[7, 0:11] = np.array([1.405, 0.887, -1.854, 0.070, -1.494, 0, 1.494,
                               -0.070, 1.854, -0.887, -1.405])
        p[8, 0:12] = np.array([1.729, 0.444, 0.722, 2.190, -2.196, 0.984,
                               -0.984, 2.196, -2.190, -0.722, -0.444, -1.729])

        out = p[n_bands - 4, 0:n_bands]

    elif phs_type == 'quad_mod':

        # Grissom's quadratic phases (unpublished)
        k = 3.4 / n_bands  # quadratic phase coefficient
        out = k * (np.arange(0, n_bands, 1) - (n_bands - 1) / 2) ** 2

    else:
        raise Exception('phase type ("{}") not recognized.'.format(phs_type))

    return out


def dz_pins(tb, sl_sep, sl_thick, g_max, g_slew, dt, b1_max=0.18,
            ptype='ex', ftype='ls', d1=0.01, d2=0.01, gambar=4258):

    r"""PINS multiband pulse design.

    Args:
        tb (float): time-bandwidth product.
        sl_sep (float): slice separation in cm.
        sl_thick (float): slice thickness in cm.
        g_max (float): max gradient amplitude in gauss/cm
        g_slew (float): max gradient sliew in gauss/cm/s
        dt (float): RF + gradient dwell time in s.
        b1_max (float): Maximum RF amplitude
        ptype (string): pulse type, 'st' (small-tip excitation), 'ex' (pi/2
            excitation pulse), 'se' (spin-echo pulse), 'inv' (inversion), or
            'sat' (pi/2 saturation pulse).
        ftype (string): type of filter to use in pulse design
        d1 (float): passband ripple level in :math:'M_0^{-1}'.
        d2 (float): stopband ripple level in :math:'M_0^{-1}'.
        gambar (float): Appropriate gyromagnetic ratio in Hz/gauss.

    Returns:
        2-element tuple containing:

        - **rf** (*array*): RF Pulse in Gauss
        - **g** (*array*): Gradient waveform in Gauss/cm

    References:
        Norris, D.G. and Koopmans, P.J. and Boyacioglu, R and Barth, M (2011).
        'Power independent of number of slices (PINS) radiofrequency Pulses
        for low-power simultaneous multislice excitation'.
        Magn. Reson. Med., 66(5):1234-1240.
    """

    kz_width = tb / sl_thick  # 1/cm, width in k-space we must go
    # calcualte number of subpulses (odd)
    n_pulses = int(2 * np.floor(np.ceil(kz_width / (1 / sl_sep)) / 2))
    # call SLR to get envelope
    rf_soft = slr.dzrf(n_pulses, tb, ptype, ftype, d1, d2)

    # design the blip trapezoid
    g_area = 1 / sl_sep / gambar
    [gz_blip, _] = trap_grad(g_area, g_max, g_slew, dt)

    # Calculate the block/hard RF pulse width based on
    b1_scaled = 2 * np.pi * gambar * b1_max * dt
    hpw = int(np.ceil(np.max(np.abs(rf_soft)) / b1_scaled))

    # interleave RF subpusles with gradient subpulses to form full pulses
    rf = np.kron(rf_soft[:-1], np.concatenate((np.ones(hpw),
                 np.zeros((np.size(gz_blip))))))
    rf = np.concatenate((rf, rf_soft[-1] * np.ones(hpw)))
    rf = rf / (np.sum(rf) * 2 * np.pi * gambar * dt) * np.sum(rf_soft)

    g = np.concatenate([np.zeros(hpw), np.squeeze(gz_blip)])
    g = np.tile(g, n_pulses - 1)
    g = np.concatenate((g, np.zeros(hpw)))

    return rf, g



import unittest
import numpy as np
import sigpy as sp
import numpy.testing as npt
import scipy.ndimage.filters as filt

from sigpy.mri import rf, linop, sim

if __name__ == '__main__':
    unittest.main()


class TestPtx(unittest.TestCase):

    @staticmethod
    def problem_2d(dim):
        img_shape = [dim, dim]
        sens_shape = [8, dim, dim]

        # target - slightly blurred circle
        x, y = np.ogrid[-img_shape[0] / 2: img_shape[0] - img_shape[0] / 2,
                        -img_shape[1] / 2: img_shape[1] - img_shape[1] / 2]
        circle = x * x + y * y <= int(img_shape[0] / 6) ** 2
        target = np.zeros(img_shape)
        target[circle] = 1
        target = filt.gaussian_filter(target, 1)
        target = target.astype(np.complex)

        sens = sim.birdcage_maps(sens_shape)

        return target, sens

    @staticmethod
    def problem_3d(dim, Nz):
        Nc = 8
        img_shape = [dim, dim, Nz]
        sens_shape = [Nc, dim, dim, Nz]

        # target - slightly blurred circle
        x, y, z = np.ogrid[-img_shape[0] / 2: img_shape[0] - img_shape[0] / 2,
                           -img_shape[1] / 2: img_shape[1] - img_shape[1] / 2,
                           -img_shape[2] / 2: img_shape[2] - img_shape[2] / 2]
        circle = x * x + y * y + z * z <= int(img_shape[0] / 5) ** 2
        target = np.zeros(img_shape)
        target[circle] = 1
        target = filt.gaussian_filter(target, 1)
        target = target.astype(np.complex)
        sens = sp.mri.sim.birdcage_maps(sens_shape)

        return target, sens

    def test_stspa_radial(self):

        target, sens = self.problem_2d(8)

        # makes dim*dim*2 trajectory
        traj = sp.mri.radial((sens.shape[1], sens.shape[1], 2),
                             target.shape, golden=True, dtype=np.float)
        # reshape to be Nt*2 trajectory
        traj = np.reshape(traj, [traj.shape[0]*traj.shape[1], 2])

        A = linop.Sense(sens, coord=traj,
                        weights=None, ishape=target.shape).H

        pulses = rf.stspa(target, sens, traj, dt=4e-6, alpha=1,
                          b0=None, st=None,
                          explicit=False, max_iter=100, tol=1E-4)

        npt.assert_array_almost_equal(A*pulses, target, 1E-3)

    def test_stspa_spiral(self):

        target, sens = self.problem_2d(8)

        fov = 0.55
        gts = 6.4e-6
        gslew = 190
        gamp = 40
        R = 1
        dx = 0.025  # in m
        # construct a trajectory
        g, k, t, s = rf.spiral_arch(fov / R, dx, gts, gslew, gamp)

        A = linop.Sense(sens, coord=k, ishape=target.shape).H

        pulses = rf.stspa(target, sens, k, dt=4e-6, alpha=1,
                          b0=None, st=None,
                          explicit=False, max_iter=100, tol=1E-4)

        npt.assert_array_almost_equal(A*pulses, target, 1E-3)

    def test_stspa_2d_explicit(self):
        target, sens = self.problem_2d(8)
        dim = target.shape[0]
        g, k1, t, s = rf.spiral_arch(0.24, dim, 4e-6, 200, 0.035)
        k1 = k1 / dim

        A = rf.PtxSpatialExplicit(sens, k1, dt=4e-6, img_shape=target.shape,
                                  b0=None)
        pulses = sp.mri.rf.stspa(target, sens, st=None, coord=k1, dt=4e-6,
                                 max_iter=100, alpha=10, tol=1E-4,
                                 phase_update_interval=200, explicit=True)

        npt.assert_array_almost_equal(A*pulses, target, 1E-3)

    def test_stspa_3d_explicit(self):
        nz = 4
        target, sens = self.problem_3d(3, nz)
        dim = target.shape[0]

        g, k1, t, s = rf.spiral_arch(0.24, dim, 4e-6, 200, 0.035)
        k1 = k1 / dim

        k1 = rf.stack_of(k1, nz, 0.1)
        A = rf.linop.PtxSpatialExplicit(sens, k1, dt=4e-6,
                                        img_shape=target.shape, b0=None)

        pulses = sp.mri.rf.stspa(target, sens, st=None,
                                 coord=k1,
                                 dt=4e-6, max_iter=30, alpha=10, tol=1E-3,
                                 phase_update_interval=200, explicit=True)

        npt.assert_array_almost_equal(A*pulses, target, 1E-3)

    def test_stspa_3d_nonexplicit(self):
        nz = 3
        target, sens = self.problem_3d(3, nz)
        dim = target.shape[0]

        g, k1, t, s = rf.spiral_arch(0.24, dim, 4e-6, 200, 0.035)
        k1 = k1 / dim

        k1 = rf.stack_of(k1, nz, 0.1)
        A = sp.mri.linop.Sense(sens, k1, weights=None, tseg=None,
                               ishape=target.shape).H

        pulses = sp.mri.rf.stspa(target, sens, st=None,
                                 coord=k1,
                                 dt=4e-6, max_iter=30, alpha=10, tol=1E-3,
                                 phase_update_interval=200, explicit=False)

        npt.assert_array_almost_equal(A*pulses, target, 1E-3)

    def test_spokes(self):

        # spokes problem definition:
        dim = 20  # size of the b1 matrix loaded
        n_spokes = 5
        fov = 20  # cm
        dx_max = 2  # cm
        gts = 4E-6
        sl_thick = 5  # slice thickness, mm
        tbw = 4
        dgdtmax = 18000  # g/cm/s
        gmax = 2  # g/cm

        _, sens = self.problem_2d(dim)
        roi = np.zeros((dim, dim))
        radius = dim//2
        cx, cy = dim//2, dim//2
        y, x = np.ogrid[-radius:radius, -radius:radius]
        index = x**2 + y**2 <= radius**2
        roi[cy-radius:cy+radius, cx-radius:cx+radius][index] = 1
        sens = sens * roi

        [pulses, g] = rf.stspk(roi, sens, n_spokes, fov, dx_max, gts, sl_thick,
                               tbw, dgdtmax, gmax, alpha=1)

        # should give the number of pulses corresponding to number of TX ch
        npt.assert_equal(np.shape(pulses)[0], np.shape(sens)[0])
        # should hit the max gradient constraint
        npt.assert_almost_equal(gmax, np.max(g), decimal=3)



import unittest
from sigpy import version

if __name__ == '__main__':
    unittest.main()


class TestVersion(unittest.TestCase):

    def test_version(self):
        assert version.__version__ == '0.1.22'



import unittest
import numpy as np
import numpy.testing as npt
from sigpy import backend, conv, util, config


if __name__ == '__main__':
    unittest.main()

dtypes = [np.float32, np.float64,
          np.complex64, np.complex128]


class TestConv(unittest.TestCase):

    def test_convolve_valid(self):
        mode = 'valid'
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for device in devices:
            xp = device.xp
            with device:
                for dtype in dtypes:
                    with self.subTest(dtype=dtype, device=device):
                        data = util.dirac([1, 3], device=device, dtype=dtype)
                        filt = xp.ones([1, 3], dtype=dtype)
                        output = backend.to_device(conv.convolve(
                            data, filt, mode=mode))
                        npt.assert_allclose(output, [[1]], atol=1e-5)

                        data = util.dirac([1, 3], device=device, dtype=dtype)
                        filt = xp.ones([1, 2], dtype=dtype)
                        output = backend.to_device(conv.convolve(
                            data, filt, mode=mode))
                        npt.assert_allclose(output, [[1, 1]], atol=1e-5)

                        data = util.dirac([1, 1, 3], device=device,
                                          dtype=dtype)
                        filt = xp.ones([2, 1, 1, 3], dtype=dtype)
                        output = backend.to_device(
                            conv.convolve(data, filt,
                                          mode=mode,
                                          multi_channel=True),
                            backend.cpu_device)
                        npt.assert_allclose(output, [[[1]],
                                                     [[1]]], atol=1e-5)

                        data = util.dirac([1, 1, 3], device=device,
                                          dtype=dtype)
                        filt = xp.ones([2, 1, 1, 3], dtype=dtype)
                        strides = [1, 2]
                        output = backend.to_device(
                            conv.convolve(data, filt,
                                          mode=mode, strides=strides,
                                          multi_channel=True),
                            backend.cpu_device)
                        npt.assert_allclose(output, [[[1]],
                                                     [[1]]], atol=1e-5)

    def test_convolve_full(self):
        mode = 'full'
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for device in devices:
            xp = device.xp
            with device:
                for dtype in dtypes:
                    with self.subTest(dtype=dtype, device=device):
                        data = util.dirac([1, 3], device=device, dtype=dtype)
                        filt = xp.ones([1, 3], dtype=dtype)
                        output = backend.to_device(conv.convolve(
                            data, filt, mode=mode))
                        npt.assert_allclose(output, [[0, 1, 1, 1, 0]],
                                            atol=1e-5)

                        data = util.dirac([1, 3], device=device, dtype=dtype)
                        filt = xp.ones([1, 2], dtype=dtype)
                        output = backend.to_device(conv.convolve(
                            data, filt, mode=mode))
                        npt.assert_allclose(output, [[0, 1, 1, 0]], atol=1e-5)

                        data = util.dirac([1, 1, 3], device=device,
                                          dtype=dtype)
                        filt = xp.ones([2, 1, 1, 3], dtype=dtype)
                        output = backend.to_device(
                            conv.convolve(data, filt,
                                          mode=mode,
                                          multi_channel=True),
                            backend.cpu_device)
                        npt.assert_allclose(output, [[[0, 1, 1, 1, 0]],
                                                     [[0, 1, 1, 1, 0]]],
                                            atol=1e-5)

                        data = util.dirac([1, 1, 3], device=device,
                                          dtype=dtype)
                        filt = xp.ones([2, 1, 1, 3], dtype=dtype)
                        strides = [1, 2]
                        output = backend.to_device(
                            conv.convolve(data, filt,
                                          mode=mode,
                                          strides=strides,
                                          multi_channel=True))
                        npt.assert_allclose(output, [[[0, 1, 0]],
                                                     [[0, 1, 0]]], atol=1e-5)

    def test_convolve_data_adjoint_valid(self):
        mode = 'valid'
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for device in devices:
            xp = device.xp
            with device:
                for dtype in dtypes:
                    with self.subTest(dtype=dtype, device=device):
                        output = xp.ones([1, 1], dtype=dtype)
                        filt = xp.ones([1, 3], dtype=dtype)
                        data_shape = [1, 3]
                        data = backend.to_device(conv.convolve_data_adjoint(
                            output, filt, data_shape, mode=mode))
                        npt.assert_allclose(data, [[1, 1, 1]], atol=1e-5)

                        output = xp.ones([1, 2], dtype=dtype)
                        filt = xp.ones([1, 2], dtype=dtype)
                        data_shape = [1, 3]
                        data = backend.to_device(conv.convolve_data_adjoint(
                            output, filt, data_shape, mode=mode))
                        npt.assert_allclose(data, [[1, 2, 1]], atol=1e-5)

                        output = xp.ones([2, 1, 1], dtype=dtype)
                        filt = xp.ones([2, 1, 1, 3], dtype=dtype)
                        data_shape = [1, 1, 3]
                        data = backend.to_device(
                            conv.convolve_data_adjoint(
                                output, filt, data_shape,
                                mode=mode,
                                multi_channel=True),
                            backend.cpu_device)
                        npt.assert_allclose(data, [[[2, 2, 2]]], atol=1e-5)

                        output = xp.ones([2, 1, 1], dtype=dtype)
                        filt = xp.ones([2, 1, 1, 3], dtype=dtype)
                        data_shape = [1, 1, 4]
                        strides = [1, 2]
                        data = backend.to_device(
                            conv.convolve_data_adjoint(
                                output, filt, data_shape,
                                mode=mode, strides=strides,
                                multi_channel=True),
                            backend.cpu_device)
                        npt.assert_allclose(data, [[[2, 2, 2, 0]]], atol=1e-5)

    def test_convolve_data_adjoint_full(self):
        mode = 'full'
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for device in devices:
            xp = device.xp
            with device:
                for dtype in dtypes:
                    with self.subTest(dtype=dtype, device=device):
                        output = xp.ones([1, 5], dtype=dtype)
                        filt = xp.ones([1, 3], dtype=dtype)
                        data_shape = [1, 3]
                        data = backend.to_device(conv.convolve_data_adjoint(
                            output, filt, data_shape, mode=mode))
                        npt.assert_allclose(data, [[3, 3, 3]], atol=1e-5)

                        output = xp.ones([1, 4], dtype=dtype)
                        filt = xp.ones([1, 2], dtype=dtype)
                        data_shape = [1, 3]
                        data = backend.to_device(
                            conv.convolve_data_adjoint(
                                output, filt, data_shape, mode=mode))
                        npt.assert_allclose(data, [[2, 2, 2]], atol=1e-5)

                        output = xp.ones([2, 1, 5], dtype=dtype)
                        filt = xp.ones([2, 1, 1, 3], dtype=dtype)
                        data_shape = [1, 1, 3]
                        data = backend.to_device(
                            conv.convolve_data_adjoint(
                                output, filt, data_shape,
                                mode=mode,
                                multi_channel=True),
                            backend.cpu_device)
                        npt.assert_allclose(data, [[[6, 6, 6]]], atol=1e-5)

                        output = xp.ones([2, 1, 5], dtype=dtype)
                        filt = xp.ones([2, 1, 1, 3], dtype=dtype)
                        data_shape = [1, 1, 8]
                        strides = [1, 2]
                        data = backend.to_device(
                            conv.convolve_data_adjoint(
                                output, filt, data_shape,
                                mode=mode, strides=strides,
                                multi_channel=True),
                            backend.cpu_device)
                        npt.assert_allclose(data, [[[4, 2, 4, 2, 4, 2, 4, 2]]],
                                            atol=1e-5)

    def test_convolve_filter_adjoint_valid(self):
        mode = 'valid'
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for device in devices:
            xp = device.xp
            with device:
                for dtype in dtypes:
                    with self.subTest(dtype=dtype, device=device):
                        data = xp.ones([1, 3], dtype=dtype)
                        output = xp.ones([1, 1], dtype=dtype)
                        filt_shape = [1, 3]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output, data, filt_shape, mode=mode))
                        npt.assert_allclose(filt, [[1, 1, 1]], atol=1e-5)

                        data = xp.ones([1, 3], dtype=dtype)
                        output = xp.ones([1, 2], dtype=dtype)
                        filt_shape = [1, 2]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output, data, filt_shape, mode=mode))
                        npt.assert_allclose(filt, [[2, 2]], atol=1e-5)

                        data = xp.ones([1, 1, 3], dtype=dtype)
                        output = xp.ones([2, 1, 1], dtype=dtype)
                        filt_shape = [2, 1, 1, 3]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output, data, filt_shape,
                                mode=mode,
                                multi_channel=True),
                            backend.cpu_device)
                        npt.assert_allclose(filt, [[[[1, 1, 1]]],
                                                   [[[1, 1, 1]]]], atol=1e-5)

                        data = xp.ones([1, 1, 4], dtype=dtype)
                        output = xp.ones([2, 1, 1], dtype=dtype)
                        filt_shape = [2, 1, 1, 3]
                        strides = [1, 2]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output, data, filt_shape,
                                mode=mode, strides=strides,
                                multi_channel=True),
                            backend.cpu_device)
                        npt.assert_allclose(filt, [[[[1, 1, 1]]],
                                                   [[[1, 1, 1]]]], atol=1e-5)

    def test_convolve_filter_adjoint_full(self):
        mode = 'full'
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for device in devices:
            xp = device.xp
            with device:
                for dtype in dtypes:
                    with self.subTest(dtype=dtype, device=device):
                        data = xp.ones([1, 3], dtype=dtype)
                        output = xp.ones([1, 5], dtype=dtype)
                        filt_shape = [1, 3]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output, data, filt_shape, mode=mode))
                        npt.assert_allclose(filt, [[3, 3, 3]], atol=1e-5)

                        data = xp.ones([1, 3], dtype=dtype)
                        output = xp.ones([1, 4], dtype=dtype)
                        filt_shape = [1, 2]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output, data, filt_shape, mode=mode))
                        npt.assert_allclose(filt, [[3, 3]], atol=1e-5)

                        data = xp.ones([1, 1, 3], dtype=dtype)
                        output = xp.ones([2, 1, 5], dtype=dtype)
                        filt_shape = [2, 1, 1, 3]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output, data, filt_shape,
                                mode=mode,
                                multi_channel=True),
                            backend.cpu_device)
                        npt.assert_allclose(filt, [[[[3, 3, 3]]],
                                                   [[[3, 3, 3]]]], atol=1e-5)

                        data = xp.ones([1, 1, 3], dtype=dtype)
                        output = xp.ones([2, 1, 3], dtype=dtype)
                        filt_shape = [2, 1, 1, 3]
                        strides = [1, 2]
                        filt = backend.to_device(
                            conv.convolve_filter_adjoint(
                                output, data, filt_shape,
                                mode=mode, strides=strides,
                                multi_channel=True),
                            backend.cpu_device)
                        npt.assert_allclose(filt, [[[[2, 1, 2]]],
                                                   [[[2, 1, 2]]]], atol=1e-5)



# -*- coding: utf-8 -*-
"""MRI waveform import/export files.
"""

import numpy as np
import struct

__all__ = ['signa', 'ge_rf_params', 'philips_rf_params']


def signa(wav, filename, scale=-1):
    """Write a binary waveform in the GE format.

    Args:
        wav (array): waveform (gradient or RF), may be complex-valued.
        filename (string): filename to write to.
        scale (float): scaling factor to apply (default = waveform's max)

    Adapted from John Pauly's RF Tools signa() MATLAB function

    """

    wmax = int('7ffe', 16)

    if not np.iscomplexobj(wav):

        if scale == -1:
            scale = 1 / np.max(np.abs(wav))

        # scale up to fit in a short integer
        wav = wav * scale * wmax

        # mask off low bit, since it would be an EOS otherwise
        wav = 2 * np.round(wav / 2)

        fid = open(filename, 'wb')

        for x in np.nditer(wav):
            fid.write(struct.pack('>h', int(x.item())))

        fid.close()

    else:

        if scale == -1:
            scale = 1 / np.max(
                (np.max(np.abs(np.real(wav))), np.max(np.abs(np.imag(wav)))))

        # scale up to fit in a short integer
        wav = wav * scale * wmax

        # mask off low bit, since it would be an EOS otherwise
        wav = 2 * np.round(wav / 2)

        fid = open(filename + '.r', 'wb')

        for x in np.nditer(wav):
            fid.write(struct.pack('>h', np.real(x)))

        fid.close()

        fid = open(filename + '.i', 'wb')

        for x in np.nditer(wav):
            fid.write(struct.pack('>h', np.imag(x)))

        fid.close()


def ge_rf_params(rf, dt=4e-6):
    """Calculate RF pulse parameters for deployment
    on a GE scanner.

    Args:
        rf (array): RF pulse samples
        dt (scalar): RF dwell time (seconds)

    Adapted from Adam Kerr's rf_save() MATLAB function

    """

    print('GE RF Pulse Parameters:')

    n = len(rf)
    rfn = rf / np.max(np.abs(rf))

    abswidth = np.sum(np.abs(rfn)) / n
    print('abswidth = ', abswidth)

    effwidth = np.sum(np.abs(rfn) ** 2) / n
    print('effwidth = ', effwidth)

    print('area = ', abswidth)

    pon = np.abs(rfn) > 0
    temp_pw = 0
    max_pw = 0
    for i in range(0, len(rfn)):
        if pon[i] == 0 & temp_pw > 0:
            max_pw = np.max(max_pw, temp_pw)
            temp_pw = 0
    max_pw = max_pw / n

    dty_cyc = np.sum(np.abs(rfn) > 0.2236) / n
    if dty_cyc < max_pw:
        dty_cyc = max_pw
    print('dtycyc = ', dty_cyc)
    print('maxpw = ', max_pw)

    max_b1 = np.max(np.abs(rf))
    print('max_b1 = ', max_b1)

    int_b1_sqr = np.sum(np.abs(rf) ** 2) * dt * 1e3
    print('int_b1_sqr = ', int_b1_sqr)

    rms_b1 = np.sqrt(np.sum(np.abs(rf) ** 2)) / n
    print('max_rms_b1 = ', rms_b1)


def philips_rf_params(rf):
    """Calculate RF pulse parameters for deployment
    on a Philips scanner.

    Args:
        rf (array): RF pulse samples (assumed real-valued)

    """

    print('Philips RF Pulse Parameters')

    n = len(rf)
    rfn = rf / np.max(np.abs(rf))

    am_c_teff = np.sum(rfn * 32767) / (32767 * n)
    print('am_c_teff = ', am_c_teff)

    am_c_trms = np.sum((rfn * 32767) ** 2) / (32767 ** 2 * n)
    print('am_c_trms = ', am_c_trms)

    am_c_tabs = np.sum(np.abs(rfn) * 32767) / (32767 * n)
    print('am_c_tabs = ', am_c_tabs)

    # assume that the isodelay point is at the peak
    am_c_sym = np.argmax(np.abs(rfn)) / n
    print('am_c_sym = ', am_c_sym)
